unit glstmainu;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, Buttons, StdCtrls,
  ComCtrls, GLScene, GLViewer, GLMaterial, GLMultiMaterialShader,
  GLHiddenLineShader, GLOutlineShader, GLObjects, Types, GLVectorGeometry,
  GLGeomObjects, GLTexture, GLCadencer, uNavCube, GLBaseClasses;

type

  { TForm1 }

  TForm1 = class(TForm)
    Button1: TButton;
    Button2: TButton;
    GLArrowLine1: TGLArrowLine;
    GLArrowLine2: TGLArrowLine;
    GLCadencer1: TGLCadencer;
    GLCamera1: TGLCamera;
    GLCube1: TGLCube;
    GLDummyCube1: TGLDummyCube;
    container: TGLDummyCube;
    GLHiddenLineShader1: TGLHiddenLineShader;
    GLLightSource1: TGLLightSource;
    GLLightSource2: TGLLightSource;
    GLOutlineShader1: TGLOutlineShader;
    GLPlane1: TGLPlane;
    GLPolygon1: TGLPolygon;
    Material: TGLMaterialLibrary;
    GLScene1: TGLScene;
    Memo1: TMemo;
    SpeedButton1: TSpeedButton;
    sv: TGLSceneViewer;
    DeepOfView: TTrackBar;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure DeepOfViewChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure GLCadencer1Progress(Sender: TObject; const deltaTime,
      newTime: Double);
    procedure SpeedButton1Click(Sender: TObject);
    procedure svMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure svMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure svMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure svMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
  private
    oldx, oldy : integer;
    vmd : boolean; // Viewer Mouse is Down
  public
    navCube : TGLNavCube;
  end;

var
  Form1: TForm1;

implementation
uses math, glgeometrybb, glvectorfileobjects;

{$R *.lfm}

{ TForm1 }

procedure TForm1.svMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  vmd := true;
  oldx := x;
  oldy := y;
end;

procedure TForm1.SpeedButton1Click(Sender: TObject);
var dc : TGLDummyCube;
    bb : THmgBoundingBox;
    fx, fy, sz : Single;
    g, k,
    min, max : TAffineVector;
    w, h : integer;
    ss : Single;
begin
  bb := container.BoundingBoxAbsolute;
  Memo1.Lines.Add(Format('%.3f / %.3f / %.3f', [bb.bbox[0].x, bb.bbox[0].y, bb.bbox[0].z]));
  Memo1.Lines.Add(Format('%.3f / %.3f / %.3f', [bb.bbox[1].x, bb.bbox[1].y, bb.bbox[1].z]));
  Memo1.Lines.Add(Format('%.3f / %.3f / %.3f', [bb.bbox[2].x, bb.bbox[2].y, bb.bbox[2].z]));
  Memo1.Lines.Add(Format('%.3f / %.3f / %.3f', [bb.bbox[3].x, bb.bbox[3].y, bb.bbox[3].z]));
  Memo1.Lines.Add(Format('%.3f / %.3f / %.3f', [bb.bbox[4].x, bb.bbox[4].y, bb.bbox[4].z]));
  Memo1.Lines.Add(Format('%.3f / %.3f / %.3f', [bb.bbox[5].x, bb.bbox[5].y, bb.bbox[5].z]));
  Memo1.Lines.Add(Format('%.3f / %.3f / %.3f', [bb.bbox[6].x, bb.bbox[6].y, bb.bbox[6].z]));
  Memo1.Lines.Add(Format('%.3f / %.3f / %.3f', [bb.bbox[7].x, bb.bbox[7].y, bb.bbox[7].z]));

  dc := TGLDummyCube.Create(GLScene1);
  sz := 0.0;
  sz := Math.Max(sz, bbMaxX(bb)-bbMinx(bb));
  sz := Math.Max(sz, bbMaxY(bb)-bbMinY(bb));
  sz := Math.Max(sz, bbMaxZ(bb)-bbMinZ(bb));
  dc.CubeSize:=sz;
  dc.Translate((bbMaxX(bb)+bbMinX(bb)) / 2,
               (bbMaxY(bb)+bbMinY(bb)) / 2,
               (bbMaxZ(bb)+bbMinZ(bb)) / 2);
  dc.VisibleAtRunTime:=true;
  Container.AddChild(dc);
  min.x := bbMinX(bb);
  min.y := bbMinY(bb);
  min.z := bbMinZ(bb);
  max.x := bbMaxX(bb);
  max.y := bbMaxY(bb);
  max.z := bbMaxZ(bb);
  k := sv.buffer.WorldToScreen(min);
  g := sv.Buffer.WorldToScreen(max);
  ss := sv.Camera.SceneScale;
  w := sv.ClientWidth;
  h := sv.ClientHeight;
  fx :=  w / (g.x - k.x);
  fx := fx * ss;
  fy := h / (g.y - k.y) * ss;
  if (fx > fy) then
   sv.Camera.SceneScale := fy
  else
   sv.Camera.SceneScale := fx;
end;

procedure TForm1.Button1Click(Sender: TObject);
var Line : TGLArrowLine;
begin
  Line := TGLArrowLine.Create(GLSCene1);
  Line.Material.MaterialLibrary := Material;
  Line.Material.LibMaterialName:='blau';
  Line.Direction.asVector := VectorCrossProduct(GLArrowLine1.Direction.asVector, GLArrowLine2.Direction.asVector);
  Line.Translate(0, 0, 0.5);
  Container.AddChild(Line);
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  if NavCube = nil then
  begin
    NavCube := TGLNavCube.CreateAsChild(Container);
    NavCube.SceneViewer := sv;
    NavCube.Camera:=GLCamera1;
    NavCube.ActiveMouse:=False;
   // NavCube.AutoRotate:=False;
    //NavCube.FPS := 30;
  end;
end;

procedure TForm1.DeepOfViewChange(Sender: TObject);
begin
  sv.Camera.DepthOfView:=DeepOfView.Position;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
end;

procedure TForm1.GLCadencer1Progress(Sender: TObject; const deltaTime,
  newTime: Double);
begin
  if not Assigned(NavCube) then exit;
  if NavCube.InactiveTime > 5 then
  begin
    if NavCube.InactiveTime < 8 then
      GLCamera1.TurnAngle := GLCamera1.TurnAngle + (NavCube.InactiveTime - 5) * deltaTime * 2
    else
      GLCamera1.TurnAngle := GLCamera1.TurnAngle + deltatime * 6;
  end;
  sv.Refresh;
  if Self.Focused then
    sv.Invalidate;
end;

procedure TForm1.svMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var dx, dy : integer;
    mp, dp : TAffineVector;
    diff : TAffineVector;
begin
  if vmd then
  begin
    mp := sv.Buffer.OrthoScreenToWorld(oldx, sv.clientheight-oldy);
    dp := sv.Buffer.OrthoScreenToWorld(x, sv.clientheight-y);
    diff := VectorSubtract(mp, dp);
    dx := oldx - x;
    dy := oldy - y;
    if ssLeft in Shift then
      GLCamera1.MoveAroundTarget(dy, dx)
    else if ssRight in Shift then
      GLCamera1.TargetObject.Translate(diff.x, diff.y, diff.z);
    oldx := x;
    oldy := y;
  end
end;

procedure TForm1.svMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  vmd := false;
end;

procedure TForm1.svMouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
var sc : single;
    mp, cp, diff : TAffineVector;
    d : TAffineVector;
    dx, dy : integer;
begin
  // Die Zoomerei passiert in der Mitte der aktuellen Ansicht. Das wollen wir nicht.
  // Wir müssen zunächst die Scene verschieben, so dass der Punkt unter der Maus in
  // der Mitte der Ansicht ist. Dann wird gezoomt und danach der Punkt, der vor dem Zoomen
  // unter der Maus stand wieder unter die Maus verschoben.
  dx := sv.ClientWidth shr 1;
  dy := sv.ClientHeight shr 1;
  if sv.Camera.CameraStyle = csOrthogonal then
  begin
    mp := sv.Buffer.OrthoScreenToWorld(MousePos.x, sv.clientheight-MousePos.y); // Punkt unter der Maus ermitteln
    cp := sv.Buffer.OrthoScreenToWorld(dx, sv.clientheight-dy);                 // aktuelles Zentrum ermitteln
  end else
  begin
    mp := sv.Buffer.ScreenToWorld(MousePos.x, sv.clientheight-MousePos.y); // Punkt unter der Maus ermitteln
    cp := sv.Buffer.ScreenToWorld(dx, sv.clientheight-dy);                 // aktuelles Zentrum ermitteln
  end;

  diff := VectorSubtract(mp, cp);                             // Differenz berechnen
  sv.Camera.TargetObject.Translate(diff.x, diff.y, diff.z);
  //sv.Camera.MoveTargetInEyeSpace(0, diff.x, -diff.y);       // Punkt unter der Maus in das Zentrum verschieben

  // Note that 1 wheel-step induces a WheelDelta of 120,
  // this code adjusts the distance to target with a 10% per wheel-step ratio
  if sv.Camera.CameraStyle = csPerspective then
    sv.Camera.AdjustDistanceToTarget(Power(1.1, -WheelDelta / 120))
  else if sv.Camera.CameraStyle = csOrthogonal then
  begin
    sc := sv.Camera.SceneScale;
    if WheelDelta > 0 then
     sc := sc * 1.1
    else
     sc := sc * 0.9;
    sv.Camera.SceneScale := sc;
  end;

  // Herausfinden, wo der Punkt, der vorher unter der Maus war jetzt ist
  if sv.Camera.CameraStyle = csOrthogonal then
  begin
    mp := sv.Buffer.OrthoScreenToWorld(dx, sv.clientheight-dy);
    cp := sv.Buffer.OrthoScreenToWorld(MousePos.x, sv.clientheight-MousePos.y);
  end else
  begin
    mp := sv.Buffer.ScreenToWorld(dx, sv.clientheight-dy);
    cp := sv.Buffer.ScreenToWorld(MousePos.x, sv.clientheight-MousePos.y);
  end;
  diff := VectorSubtract(mp, cp);
  sv.Camera.TargetObject.Translate(diff.x, diff.y, diff.z);
end;

end.

